import time
import unittest
from DLMS_SPODES.types import cdt
from DLMS_SPODES.cosem_interface_classes import collection
from DLMSAdapter.xml_ import xml50
from DLMSCommunicationProfile.HDLC.hdlc import HDLCParameters, HDLC
from src.DLMS_SPODES_client.client import Client, Network, Serial, IDFactory, RS485, BLEKPZ
from src.DLMS_SPODES_client.servers import TransactionServer
from src.DLMS_SPODES_client import task


task.get_adapter(adapter := xml50)


port1 = "COM6"
port2 = "COM5"

id_ = collection.ID(
    man=b'XXX',
    f_id=collection.ParameterValue(b'1234567', cdt.OctetString(bytearray(b'M2M-1')).encoding),
    f_ver=collection.ParameterValue(b'1234560', cdt.OctetString(bytearray(b'1.4.2')).encoding)
)

c_Serial_LOW = Client(
    # secret="30 30 30 30 30 30 30 30",    # for 101, 102, 103, 104 before ver 1.0
    id_="public",
    response_timeout=3,
    media=Serial(
        port=port2))
"""Serial LOW"""
c_Serial_Reader = Client(
    com_profile=HDLC(parameters=HDLCParameters(device_address=0)),
    id_="reader",
    secret="30 30 30 30 30 30 30 30",    # for 101, 102, 103, 104 before ver 1.0
    SAP=0x20,
    response_timeout=3,
    media=Serial(
        port=port1))
"""Serial LOW"""
c_Serial_Reader.m_id.set(1)  # for KPZ
fobos1_Serial_Reader = Client(
    com_profile=HDLC(parameters=HDLCParameters(device_address=0)),
    m_id=1,
    secret=b"Reader",
    SAP=0x20,
    response_timeout=3,
    addr_size=1,
    media=Serial(
        port="COM3"))
"""фобос1 reader"""
c_Serial_HIGH = Client(
    id_="c_Serial_HIGH",
    # secret="30 30 30 30 30 30 30 30",    # for 101, 102, 103, 104 before ver 1.0
    secret="30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30",  # for KPZ
    SAP=0x30,
    response_timeout=30,
    media=Serial(
        port=port1))
"""Serial HIGH"""
c_Serial_HIGH.m_id.set(2)  # for KPZ
c_Serial_HIGH2 = Client(
    id_="c_Serial_HIGH2",
    # secret="30 30 30 30 30 30 30 30",    # for 101, 102, 103, 104 before ver 1.0
    secret="30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30",  # for KPZ
    SAP=0x30,
    response_timeout=20,
    media=Serial(
        port=port2))
"""Serial HIGH2"""
c_Serial_HIGH2.m_id.set(2)  # for KPZ
c_Net_LOW = Client(
    media=Network(
        host="127.0.0.1",
        port='10000'),
    response_timeout=3)
"""Network LOW"""
c_BLE_LOW = Client(
    media=BLEKPZ(
        addr="A0:6C:65:53:7D:86"),
    response_timeout=3)
"""BLEKPZ LOW"""
c_BLE_HIGH = Client(
    secret="30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30",  # for KPZ
    SAP=0x30,
    media=BLEKPZ(
        addr="A0:6C:65:53:7D:86"),
    response_timeout=3)
c_BLE_HIGH.m_id.set(2)  # for KPZ
"""BLEKPZ High"""


class TestType(unittest.TestCase):
    def test_create_Client(self):
        id_factory = IDFactory("d")
        c1 = Client(id_=id_factory.create())
        c2 = Client(id_="d2", del_cb=id_factory.remove)
        del c2
        Client(id_=id_factory.create())
        Client()
        c3 = Client()
        c3.com_profile = HDLC()
        print(c1.id, c3.id, id_factory.value)

    def test_lowest_connect(self):
        t_server = TransactionServer(
            clients=[c_Serial_LOW],
            tsk=task.Dummy()
        )
        t_server.start()
        while not t_server.results.is_complete():
            time.sleep(1)
        print(c_Serial_LOW.objects)

    def test_reader_connect(self):
        t_server = TransactionServer(
            clients=[c_Serial_Reader],
            tsk=task.Dummy()
        )
        t_server.start()
        while not t_server.results.is_complete():
            time.sleep(1)
        print(c_Serial_LOW.objects)

    def test_fobos1_connect(self):
        t_server = TransactionServer(
            clients=[fobos1_Serial_Reader],
            tsk=task.Dummy()
        )
        t_server.start()
        while not t_server.results.is_complete():
            time.sleep(1)
        print(fobos1_Serial_Reader.objects)

    def test_lowest_network_connect(self):
        t_server = TransactionServer(
            clients=[c_Net_LOW],
            tsk=task.Dummy()
        )
        c_Net_LOW.com_profile.parameters.device_address = None
        t_server.start()
        while not t_server.results.is_complete():
            time.sleep(1)
        print(c_Net_LOW.objects)

    def test_lowest_KPZBLE_connect(self):
        t_server = TransactionServer(
            clients=[c_BLE_LOW],
            tsk=task.Dummy()
        )
        t_server.start()
        while not t_server.results.is_complete():
            time.sleep(1)
        print(c_BLE_LOW.objects)

    def test_high_connect(self):
        c_Serial_HIGH.response_timeout = 10
        c_Serial_HIGH2.response_timeout = 10
        t_server = TransactionServer(
            clients=[
                # c_Serial_HIGH,
                c_Serial_HIGH2
            ],
            tsk=task.Dummy()
        )
        t_server.start()
        while not t_server.results.is_complete():
            time.sleep(1)
        print(c_Serial_HIGH.objects)

    def test_SetLocalTime(self):
        t_server = TransactionServer(
            clients=[c := Client(
                secret="30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30",
                media=Serial(
                    port=port2),
                response_timeout=3)],
            tsk=task.SetLocalTime()
        )
        t_server.start()
        c.m_id.set(2)
        c.SAP.set(0x30)  # for KPZ
        while not t_server.results.is_complete():
            time.sleep(1)
        print(c.objects)

    def test_Loop(self):
        def foo(res):
            return res == cdt.Long(4)

        t_server = TransactionServer(
            clients=[
                client := Client(
                    secret="30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30",
                    addr_size=-1)
            ],
            tsk=task.Loop(
                task=task.ReadAttribute(
                    ln="0.0.1.0.0.255",
                    index=3),
                # func=lambda res: res == cdt.Long(4),
                func=foo,
                delay=2,
                attempt_amount=5
            )
        )
        client.m_id.set(2)
        client.SAP.set(0x30)  # for KPZ

        # client.device_address.set(0)
        client.media = Serial(
            port=port1
        )
        t_server.start()
        while not t_server.results.is_complete():
            time.sleep(1)
        print("stop")

    def test_simple_read(self):
        t_server = TransactionServer(
            clients=[
                client := Client(
                    addr_size=-1)
            ],
            # tsk=task.ReadAttribute(
            #     ln="0.0.1.0.0.255",
            #     index=2)
            tsk=task.ReadAttribute(
                ln="0.0.1.0.0.255",
                index=2)
        )
        # client.m_id.set(2)
        # client.SAP.set(0x30)  # for KPZ
        client.media = Serial(
            port=port1
        )
        t_server.start()
        while not t_server.results.is_complete():
            time.sleep(1)
        print("stop")

    def test_InitType(self):
        client = c_Serial_Reader
        client2 = c_Serial_LOW
        t_server = TransactionServer(
            clients=[
                client,
                client2
            ],
            tsk=task.init_type
        )
        t_server.start()
        while not t_server.results.is_complete():
            time.sleep(1)
            print(f"{t_server.results.is_complete()=}")
        time.sleep(3)
        print(t_server.results.ok_results)

    def test_async_network(self):
        client = Client(
            com_profile=HDLC(parameters=HDLCParameters(device_address=0)),
            secret="30 30 30 30 30 30 30 30",
            addr_size=1,
            conformance="010000000001111000011101",
            media=Serial(port="COM3")
        )
        type_ = "4d324d5f31"
        ver = "1.5.7"
        man = b"KPZ"
        # client.objects = collection.get(
        #     m=man,
        #     t=cdt.OctetString(type_),
        #     ver=AppVersion.from_str(ver))
        client.com_profile.parameters.device_address = 0
        client2 = Client(
            com_profile=HDLC(parameters=HDLCParameters(device_address=0)),
            secret="00 00 00 00 00 00 00 00",
            addr_size=1,
            conformance="010000000001111000011101",
            media=Serial(port=port1)
        )
        client2.m_id.set(1)
        client3 = Client(
            com_profile=HDLC(parameters=HDLCParameters(device_address=0)),
            secret="00 00 00 00 00 00 00 00",
            addr_size=1,
            conformance="010000000001111000011101",
            media=Network(
                host="127.0.0.1",
                port="10000")
        )
        client3.m_id.set(0)
        t_server = TransactionServer(
            clients=[client3],
            # clients=[client, client2, client3],
            tsk=(s := task.Sequence(
                task.init_type,
                task.WriteAttribute("0.0.1.0.0.255", 3, "100"),
                task.ReadAttribute("0.0.1.0.0.255", 2),
                task.WriteTime()
            )))
        t_server.start()
        # s.append(task.InitType())
        # print(f"{t_server.is_complete()=}")
        # time.sleep(1)
        # print(f"{t_server.is_complete()=}")
        # time.sleep(1)
        # print(f"{t_server.is_complete()=}")
        # t_s2 = TransactionServer2(
        #     clients=[client2, client3],
        #     exchanges=(tasks.ReadAttribute("0.0.42.0.0.255", 2),))
        # t_se.start()
        time.sleep(.3)
        for r in t_server.results:
            print(F'{r.complete=}')
        print("end")
        print(f"{t_server.results.is_complete()=}")
        time.sleep(3)
        print(t_server.results.ok_results)

    def test_server_stop(self):
        client = Client(
            com_profile=HDLC(parameters=HDLCParameters(device_address=0)),
            secret="30 30 30 30 30 30 30 30",
            addr_size=1,
            conformance="010000000001111000011101",
            response_timeout=3)
        client.SAP.set(0x20)
        client.m_id.set(1)
        client.media = Serial(
            port=port1
        )
        t_server = TransactionServer(
            clients=[client],
            tsk=(s := task.Sequence(
                task.init_type,
                # task.WriteAttribute("0.0.1.0.0.255", 3, "100"),
                # task.ReadAttribute("1.0.99.1.0.255", 2),
                task.ReadAttribute("0.0.99.98.2.255", 2),
                task.ReadAttribute("0.0.99.98.3.255", 2),
                # task.WriteTime()
            )))
        t_server.start()

        time.sleep(.3)
        for r in t_server.results:
            print(F'{r.complete=}')
        print(f"{t_server.results.is_complete()=}")
        # time.sleep(4)
        # t_server.abort()
        # time.sleep(1)
        # t_server.abort()
        print(t_server.results.ok_results)
        while not t_server.results.is_complete():
            time.sleep(1)
        a = t_server.results[0]
        print(t_server)

    def test_write(self):
        client = Client(
            com_profile=HDLC(parameters=HDLCParameters(device_address=0)),
            secret="30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30",
            addr_size=1,
            conformance="010000000001111000011101",
            media=Serial(port1))
        # type_ = "4d324d5f31"
        # ver = "1.5.7"
        # man = b"KPZ"
        # client.objects = collection.get(
        #     m=man,
        #     t=cdt.OctetString(type_),
        #     ver=AppVersion.from_str(ver))
        client.SAP.set(0x30)
        client.m_id.set(2)
        t_server = TransactionServer(
            clients=[client],
            tsk=(s := task.Sequence(
                task.init_type,
                task.WriteAttribute("0.0.1.0.0.255", 3, "100"),
            )))
        t_server.start()

        time.sleep(.3)
        for r in t_server.results:
            print(F'{r.complete=}')
        print(f"{t_server.results.is_complete()=}")
        # time.sleep(4)
        # t_server.abort()
        # time.sleep(1)
        # t_server.abort()
        print(t_server.results.ok_results)
        while not t_server.results.is_complete():
            time.sleep(1)
        a = t_server.results[0]
        print(t_server)

    def test_firmware_update(self):
        t_server = TransactionServer(
            clients=[c_Serial_HIGH2],
            tsk=task.UpdateFirmware())
        t_server.start()

        while not t_server.results.is_complete():
            time.sleep(1)
            print(f"{t_server.results.is_complete()=}")
        a = t_server.results[0]
        print(t_server)

    def test_read_association(self):
        client = Client(
            secret="30 30 30 30 30 30 30 30",
            conformance="010000000001111000011101",
            media=Serial(
                port=port1))
        # client.SAP.set(0x30)  # for KPZ
        # client.m_id.set(2)  # for KPZ
        client.m_id.set(1)  # for 101, 102, 103, 104 before ver 1.0
        t_server = TransactionServer(
            clients=[client],
            tsk=task.ReadAttribute(
                ln='0.0.40.0.1.255',
                index=2
            ))
        t_server.start()

        while not t_server.results.is_complete():
            time.sleep(1)
        print(f"{t_server.results.is_complete()=}")

    def test_OSI(self):
        from src.DLMS_SPODES_client.client import OSI

        level = OSI.PHYSICAL | OSI.DATA_LINK
        print(level)
        print(OSI.PHYSICAL not in level)
        level |= OSI.APPLICATION
        print(level)
        level -= OSI.APPLICATION
        print(level)

    def test_CreateCollection(self):
        client = Client(
            com_profile=HDLC(parameters=HDLCParameters(device_address=0)),
            SAP=0x30,
            secret="30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30",
            addr_size=1,
            media=Serial(
              port="COM3")
        )
        client.m_id.set(2)
        t_server = TransactionServer(
            clients=[client],
            tsk=task.TestAll()
        )
        t_server.start()
        while not t_server.results.is_complete():
            time.sleep(1)
            print(f"{t_server.results.is_complete()=}")
        a = t_server.results[0]
        print(t_server)

    def test_universal(self):
        client = Client(
            com_profile=HDLC(parameters=HDLCParameters(device_address=0)),
            universal=True,
            SAP=0x30,
            secret="30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30",
            addr_size=1,
            m_id=2,
            media=Serial(
              port="COM3")
        )
        client.objects = collection.Collection()
        client.objects.LDN.set_attr(2, b'\x09\x0fXXX01234567890123')
        print(client.objects, client.objects.LDN.value)
        t_server = TransactionServer(
            clients=[client],
            tsk=task.Dummy()
        )
        t_server.start()
        while not t_server.results.is_complete():
            time.sleep(1)
            print(f"{t_server.results.is_complete()=}")
        a = t_server.results[0]
        print(t_server)

    def test_IDSAP(self):
        value = task.IDSAP(
            id=id_,
            sap=task.enums.ClientSAP(0x10)
        )
        value2 = task.IDSAP(
            id=id_,
            sap=task.enums.ClientSAP(0x30)
        )
        print(hash(value), hash(value2))

    def test_KPZ_old(self):
        c = Client(
            com_profile=HDLC(),
            SAP=0x00,
            secret="30 30 30 30 30 30 30 30",    # for 101, 102, 103, 104 before ver 1.0
            id_="public",
            response_timeout=3,
            media=Serial(
                port=port1))
        """Serial LOW"""
        c.m_id.set(1)
        t_server = TransactionServer(
            clients=[c],
            tsk=task.Dummy()
        )
        t_server.start()
        while not t_server.results.is_complete():
            time.sleep(1)
        self.assertIsNotNone(c.objects, "empty collection")
        print(c.objects)
